!
!     MeanDark     Mean of the set of dark frames.
!     Copyright (C) 1998-4  Filip Hroch, Masaryk University, Brno, CZ
!

       Program MDark
!
! Experimental version 0.2 F90
!
! Z nekolika dark-framu udela prumerny snimek.
! Zadna velka veda. Jen secte pixely v jendotlivych 
! snimcich a podeli poctem
!
! $Id: mdark.f90,v 1.4 2005/02/21 16:47:50 hroch Exp $

         implicit none

! Parameters:

!    Version:
       Character(len=80),parameter :: version = 'MeanDark v. 0.0.2 - F90'

! Initial parameters.
!    Output image name:
       Character(len=256) :: namedark='meandark.fits'

!    Minimum of values
       Real :: minvalue = 0.0

!    Maximum of values
       Real :: maxvalue = 65535.0

!-------

       Character(len=256) :: sb
       Character(len=80) :: buf,dateobs,masterdateobs

       Logical :: simple,extend,anyf,exist
       Integer :: i,n, nobs, maxobs
       Integer :: istat, ostat ,bitpix, naxis, naxes(2)
       Integer :: BlockSize,pcount,gcount
       character(len=80),allocatable :: alldate(:), newdate(:)

! image dimensions
       Integer ::  nx = -1, ny = -1 
! image
       Real, allocatable :: DARK(:,:), SIG(:,:), IMAGE(:,:)
       Real :: bscale, bzero, s, tempr, temperature, stemperature
       real :: expr,exptime,sexptime

!------

       write(*,*) trim(version)

       istat = 0

       n = 0
! nacteni jmena souboru
       Do
         Do
           Read(*,'(A)',END=20,ERR=666) sb
           if( sb /= ' ' ) exit              ! blank line
         enddo
         write(*,*) trim(sb),':'

         istat = 0
! input initialisation of the current file
         Call ftopen(25,sb,0,BlockSize,istat)
         Call ftghpr(25,2,simple,bitpix,naxis,naxes,pcount,gcount,extend,istat)
         if( istat /= 0 ) goto 666
         CALL ftgkye(25,'BSCALE',bscale,buf,istat)
         CALL ftgkye(25,'BZERO',bzero,buf,istat)
!        BSCALE, BZERO not presented
         if( istat /= 0 ) then
            istat = 0
            bscale = 1.0
            bzero = 0.0
         endif
         CALL ftgkye(25,'TEMPERAT',tempr,buf,istat)
         CALL ftgkye(25,'EXPTIME',expr,buf,istat)
         CALL ftgkys(25,'DATE-OBS',dateobs,buf,istat)         
 
! first image - internal buffers initialisation
         if( nx < 0 .and. ny < 0 )then
            nx = naxes(1)
            ny = naxes(2)
            Allocate( IMAGE(nx,ny) )
            Allocate(  DARK(nx,ny) ) 
            Allocate(   SIG(nx,ny) ) 
            DARK = 0.0 
            SIG = 0.0
            temperature = 0.0
            stemperature = 0.0
            exptime = 0.0
            sexptime = 0.0
            nobs = 0
            maxobs = 2
            allocate(alldate(maxobs))
            masterdateobs = dateobs

         else       ! checking image dimensions
            if( nx /= naxes(1) .or. ny /= naxes(2) ) &
              stop 'Different dimensions of different images.'
         endif 

! image - data input
         n = n + 1
         call FTG2DE(25,1,minvalue,nx,nx,ny,IMAGE,anyf,istat)
         if( istat /= 0 ) goto 666
         call ftclos(25,istat)
         if( istat /= 0 ) goto 666

! compute sumation
         DARK = DARK + IMAGE
         SIG  = SIG  + IMAGE**2
         temperature = temperature + tempr
         stemperature = stemperature + tempr**2
         exptime = exptime + expr
         sexptime = sexptime + expr**2

         write(buf,'(a, F7.1)') trim(dateobs),tempr
         nobs = nobs + 1
         if( nobs > maxobs ) then
            allocate(newdate(maxobs))
            newdate = alldate
            deallocate(alldate)
            maxobs = 2*maxobs
            allocate(alldate(maxobs))
            alldate(1:maxobs/2) = newdate
            deallocate(newdate)
         endif
         alldate(nobs) = buf

       enddo

 20    continue

        if( n == 0 ) stop 'No input image.'

! prumerny obrazek
       DARK = DARK / n
       if( n == 1 )then
         SIG = huge(1.0)
       else
          where( SIG - n*DARK**2 > 0.0 )
             SIG = sqrt(SIG - n*DARK**2)/(n - 1.0)
          elsewhere
             SIG = -1.0
          endwhere
!         SIG = sqrt(SIG - n*DARK**2)/(n - 1.0)
       endif

! prumerna teplota
       temperature = temperature / n
       if( n == 1 ) then
          stemperature = huge(1.0)
       else
          if( stemperature - n*temperature**2 > 0.0 ) then
             stemperature = sqrt(stemperature - n*temperature**2)/(n - 1.0)
          else
             stemperature = -1.0
          endif
       end if

! prumerna expozicni doba
       exptime = exptime/n
       if( n == 1 ) then
          sexptime = huge(1.0)
       else
          if( sexptime - n*exptime**2 > 0.0 ) then
             sexptime = sqrt(sexptime - n*exptime**2)/(n - 1.0)
          else
             sexptime = 0.0
          endif
       end if

! computation of the mean of the standart deviation
       s = sum(SIG,mask=SIG>0) / count(SIG>0)

! range checking
!      where( DARK < minvalue )
!        DARK = minvalue
!      endwhere
!      where( DARK > maxvalue )
!        DARK = maxvalue
!      endwhere

       ostat = 0
       Inquire(file=namedark,exist=exist)
       if( exist ) then
         open(26,file=namedark)
         close(26,status='DELETE')
       endif
! Vytvoreni vystupniho obrazku
       Call ftinit(26,namedark,1,ostat)
! zapsani informace do hlavicky
       naxis = 2
       naxes = (/nx, ny/)
       Call ftphps(26,bitpix,naxis,naxes,ostat)
       Call ftpkys(26,'OBJECT','MEAN-DARK FRAME','',ostat)
       Call ftpkye(26,'BSCALE',bscale,10,'',ostat)
       Call ftpkye(26,'BZERO',bzero,10,'',ostat)
       call ftpkye(26,'EXPTIME',exptime,10,'Average of exposure times (seconds)',ostat)
       call ftpkye(26,'TEMPERAT',temperature,10,'Average of camera temperatures (C)',&
            ostat)
       call ftpkys(26,'DATE-OBS',masterdateobs,'UTC of the first on input',istat)
       write(buf,'(A,F6.1,A,F4.1)') 'Averaged temperature =',temperature, &
            '+-',stemperature
       Call ftpcom(26,buf,ostat)
       write(buf,'(A,F7.1,A,F4.1)') 'Averaged exposure time =',exptime, &
            '+-',sexptime
       Call ftpcom(26,buf,ostat)
       write(buf,'(a,f10.3)') 'Mean of sigmas =',s
       Call ftpcom(26,buf,ostat)

       if( nobs > 0 ) then
          write(buf,'(a,i5,a)') 'The image is a result of averadge of ',nobs,'  exposure(s).'
          call ftpcom(26,buf,ostat)
          call ftpcom(26,'The time of start and the temperature for each image used:',ostat)
          do i = 1, nobs
             call ftpcom(26,alldate(i),ostat)
          enddo
          call ftpcom(26,'All times are derived from DATE-OBS and TEMPERATURE keywords.',ostat)
       endif

       if( ostat /= 0 ) goto 666

! zapsani  dat
       Call ftp2de(26,1,nx,nx,ny,DARK,ostat)

! Zavreni souboru
       CALL ftclos(26,ostat)
       if( ostat /= 0 ) goto 666


! tisk informaci o vyslednem obrazku
       write(*,*) 'Output image: ',trim(namedark)
       write(*,*) 'Dimension: ',nx,'x',ny, & 
              '    Mean of sigmas=',s
!       write(*,*) 'Bye.'

       stop 'Bye.'

666    continue 

       if( istat +  ostat /= 0 )then
        write(*,*) 'The following error(s) has been occured:'
        do
          Call ftgmsg(buf)
          if( buf == ' ' ) exit
          write(*,*) trim(buf)
        enddo
      endif

      stop 'Crached...'

      End       

